/****************************************************************************
 * GamePlay.c
 *
 *		Overall Game routines
 *
 * 2/28/95
 *		Updated to compile using newer Think C / Metrowerks compiler
 *
 * 2/7/94
 *		Updated to work on 12 inch (512x384) monitors and 
 *		to center properly on monitors larger than 640 x 480
 *
 * 4/28/93 - 5/93
 *		Original Code Written
 ****************************************************************************/
 
#define SPRITEMAIN	1
#include "CToast.h"
#include "math.h"
#include <LowMem.h>

// Game Initialization
//
void InitializeGame(WindowPtr theWin)
{
	short			i;
	OSErr			oe;
	GDHandle		curDevice;

	// Fade out the screen (comment this out if you want to use debugger)
	GammaFadeOut();
	curDevice = GetGDevice();

	// Get Video Ptr and Dimensions
	gVideoRowBytes = (*(*curDevice)->gdPMap)->rowBytes & 0x3FFF;
	gVideoMem = (*(*curDevice)->gdPMap)->baseAddr;

	gUnivWidth = 640*6;		// 3840
	gUnivHeight = 480*8;	// 3840

	SetRect(&gUnivRect,0,0,gUnivWidth,gUnivHeight);
	gOffsetPt.v = 0;
	gOffsetPt.h = 0;

	// Choose between 512x384 and 640x480 modes
	//
	if (qd.screenBits.bounds.bottom - qd.screenBits.bounds.top < 480 ||
		qd.screenBits.bounds.right - qd.screenBits.bounds.left < 640) 
	{
		g12InchMode = true;
		gVisPlayWidth=512;
		gVisPlayHeight=384;
		gGraceTop = 128;
		gGraceLeft = 128;
	}
	else {
		gVisPlayWidth=640;
		gVisPlayHeight=480;
		gGraceTop = 160;
		gGraceLeft = 160;
	}
	gGraceRight = gVisPlayWidth-gGraceLeft;
	gGraceBottom = gVisPlayHeight-gGraceTop;

	SetRect(&gOffscreenRect,0,0,gVisPlayWidth+SafeWidth*2,gVisPlayHeight+SafeWidth*2);
	gVidOffset.h = ((qd.screenBits.bounds.right - qd.screenBits.bounds.left) - gVisPlayWidth)/2;
	gVidOffset.v = ((qd.screenBits.bounds.bottom - qd.screenBits.bounds.top) - gVisPlayHeight)/2;
	
	// Calc the center point
	gCenterP.h = (gOffscreenRect.right - gOffscreenRect.left) / 2;
	gCenterP.v = (gOffscreenRect.bottom - gOffscreenRect.top) / 2;

	// Allocate an offscreen pixel map
	//
	if ((oe = NewGWorld(&gOffScreen,0,&gOffscreenRect,NULL,NULL,0)) != noErr) {
		DebugStr("\pOy!");
		return;
	}
	gPixMap = GetGWorldPixMap(gOffScreen);
	LockPixels(gPixMap);

	gScreenRowBytes = (*gPixMap)->rowBytes & 0x3FFF;
	gScreenMem = (*gPixMap)->baseAddr;

	gPlayRect = gOffscreenRect;
	gPlayRect.right -= SafeWidth;
	gPlayRect.bottom -= SafeHeight;

	// Load User Preferences - Sound Volume, Keys
	LoadPreferences();	// Do this before loading sounds - has volume setting

	// Load the Sounds
	InitSounds();

	// Hide Menubar
	HideMenuBar(theWin); // 2/28/95 JAB
	// oldMBarHeight = MBarHeight;
	// MBarHeight = 0;
	// RectRgn(theWin->visRgn,&qd.screenBits.bounds);

	// Paint Startup Screen, Fade In
	PaintRect(&theWin->portRect);
	StartupScreen();
	GammaFadeIn();

	// Draw Status Bar - used to show initialization progress
	// DrawStatusBar(0,100);

	// Randomize the Random Number Generator
	GetDateTime((unsigned long *) &qd.randSeed);

	// Start playing the startup sound
	PlaySound(S_Startup, 4);

	// Calc dimensions of play field
	gPlayWidth = gPlayRect.right - gPlayRect.left;
	gPlayHeight = gPlayRect.bottom - gPlayRect.top;

	sTable = (SpriteInstance *) NewPtrClear(sizeof(SpriteInstance) * MaxSprites);
	if (sTable == NULL)
		DebugStr("\pOut of Mem");

	// Load in Sprites
	//
	for (i = 0; i < NbrSprites; ++i) {
		// DrawStatusBar(i,NbrSprites+1);

	if (sDef[i].class == SP_Large || sDef[i].class == SP_Med) {
			if (sDef[i].firstIconID) {
				sDef[i].lSprite = GetAsset('LSPR',sDef[i].firstIconID);
				MoveHHi(sDef[i].lSprite);
				HLock(sDef[i].lSprite);
				sDef[i].colorMaps = *sDef[i].lSprite;
				sDef[i].maskMaps = NULL;
			}
		}
	}
	// DrawStatusBar(NbrSprites,NbrSprites+1);

	InitRadar();

	// Load Top Scores
	LoadTopScores();
	// DrawStatusBar(NbrSprites+1,NbrSprites+1);
	
	SetPort(theWin);

	{
		long	dmy;
		Delay(3*60L,&dmy);
	}
}

// CleanUp before quitting
//
void CleanUp(void)
{
	EndSounds();
	ShowCursor();
	RestoreMenuBar();
	// MBarHeight = oldMBarHeight;
	CloseAssetFile(gAssetFile);
}


void BeginGame(void)
{
	short	i;
	CGrafPtr		curPort;
	GDHandle		curDevice;
	short			eMask;

	register		SpriteInstance	*sp;

	gMaxSprite = 0;
	gSpriteCnt = 0;
	spriteIDCtr = 1L;
	gSparkCnt = 0;
	gAsteroidCnt = 0;
	gRemainingShips = 3;
	gGameLevel = 0;
	gGameScore = 0L;
	gShieldPower = MaxShieldPower;

	gOffsetPt.v = 0;
	gOffsetPt.h = 0;
	gOldOffset = gOffsetPt;

	GetGWorld(&curPort,&curDevice);

	SetGWorld(gOffScreen,NULL);

	TextFont(geneva);
	TextSize(9);
	TextFace(0);
	TextMode(notSrcCopy);
	PaintRect(&gOffScreen->portRect);

	// DisplayPicture(BackgroundPICT,-1,-1);

	DrawRadarCircle();
	// DisplayPicture(RadarPICT, SafeWidth+gVisPlayWidth-128,SafeHeight+gVisPlayHeight-128);

	SetGWorld(curPort,curDevice);

	InitStatusDisplay();	// Display sprites should be at the lowest level

	// NewDebugDisplay();		// Debug Sprites Next

	InitStars();

	InitLevel();			// Asteroid Chances

	if (!gClientMode)
		NewShip();				// Ship

	InitGameSound();

	GammaFadeOut();
	MyCopyBits();
	GammaFadeIn();

	// Draw New Sprites
	for (i = 0,sp = sTable; i < gMaxSprite; ++i,++sp)
		(*sDef[sp->type].drawFunc)(sp);

	eMask = everyEvent;
	eMask &= ~keyUpMask;
	eMask &= ~keyDownMask;
	eMask &= ~autoKeyMask;
	SetEventMask(eMask);

	gGameState = GS_Play;
	InitGameEvent();
}

void EndGame(void)
{
	short			eMask;
	CGrafPtr		curPort;
	GDHandle		curDevice;
	static StringPtr promptStr = "\pGame Over";
	SpriteInstance	*sp;
	short			i;
	long			t;
	Rect			r;
	EventRecord		dmyEvent;

	EndGameEvent();

	i = gMaxSprite;
	sp = &sTable[i-1];
	while (i--) {
		if (sp->active)
			(*sDef[sp->type].eraseFunc)(sp);
		--sp;
	}

	GetGWorld(&curPort,&curDevice);

	SetGWorld(gOffScreen,NULL);

	TextFont(geneva);
	TextSize(12);
	TextFace(bold);
	TextMode(srcBic);
	CenterString(promptStr);
	DrawString(promptStr);

	SetGWorld(curPort,curDevice);
	SetRect(&r,SafeWidth,SafeHeight+StatusHeight,SafeWidth+gVisPlayWidth,SafeHeight+gVisPlayHeight);
	MyCopyRect(&r);


	eMask = everyEvent;
	eMask &= ~keyUpMask;
	SetEventMask(eMask);
	FlushEvents(everyEvent, 0);

	t = TickCount();

	IntegrateScore(gGameScore, gGameLevel);

	gGameState = GS_Attract;
	while (TickCount() - t < 180L &&
			!OSEventAvail(everyEvent, &dmyEvent))
			;
}


// Create a new batch of asteroids
void InitLevel(void)
{
	gMinAsteroids = gGameLevel*ExtraAsteroidsPerLevel + MinAsteroidsPerLevel;
	if (gMinAsteroids > MaxAsteroids)
		gMinAsteroids = MaxAsteroids;

	gScoreMultiply = 1;

	gBadGuyChance = MaxBadGuyChance - (BadGuyLevelDecrement*gGameLevel);

	if (gBadGuyChance < MinBadGuyChance)
		gBadGuyChance = MinBadGuyChance;

	gYummyChance = NormalYummyChance;
	gToasterChance = NormalToasterChance;
}

// Level was completed successfully
void EndLevel(void)
{
	++gGameLevel;
	AddScore(LevelBonusScore*gGameLevel);
	PlaySound(S_LevelCompletion, 4);
}


void GenNewEvents(void)
{
	short	n;
	if (gAsteroidCnt < gMinAsteroids)
		NewAsteroid(MyRandom(3));

	// Random Events - bad guys, yummies
	if (MyRandom(gBadGuyChance) == 0) {	// 1200 Random Bad Guys
		if (MyRandom(30) == 0)			// Every 30 minutes
			n = MyRandom(7);			// Lucky 7!
		else
			n = 1;
		while (n--) {
			switch (MyRandom(4+(gGameLevel>2? 2 : gGameLevel))) {
			case 0:
				NewCoffeePot();
				break;
			case 1:
				NewSaucer();
				break;
			case 2:
			case 3:
				NewCroissant();
				break;
			default:
				NewCube();
				break;
				
			}
		}
	}

	// Random Yummies
	if (MyRandom(gYummyChance) == 0)	// Random Yummys
		NewYummy();
	
	if (gGameLevel && MyRandom(gToasterChance) == 0 && gToasterCount < 3)
		NewToaster();

	if (gGameLevel > 1 && MyRandom(gToasterChance) == 0 && gMixerCount < 3)
		NewMixer();

}

// Main Game Loop
//
void MainGameLoop(WindowPtr theWin)
{
	register SpriteInstance	*sp;
	register short	i,n;
	register long	t;
	static EventRecord	dmyEvent;
	Rect	r;
#if DEBUGGING
	void	IdleTest();				// Debugging Function used to determine CPU load for profiler
#endif

#if __option(profile)			// 6/15 Optional profiling support
	_profile = 1;
#endif

	SetPort(theWin);

	do {
		t = TickCount();
	
		// If we're ready for next frame of animation
		//
		if (t - gLastDispTime >= gGameClockTicks) {
			//  Update the game clock
			++gGameClock;
			gLastDispTime = t;
	
			// Compact Sprite Table if necessary
			if (gMaxSprite > 20 && gMaxSprite > (gSpriteCnt<<1)) {
				for (i = n = 0; i < gMaxSprite; ++i) {
					if (sTable[i].active) {
						sTable[n] = sTable[i];
						if (gShip == &sTable[i])
							gShip = &sTable[n];
						++n;
					}
				}
				gMaxSprite = n;
			}
	

			// Erase all Sprites from offscreen gworld
			// Each sprite has a pointer to an erase function
			//
			i = gMaxSprite;
			sp = &sTable[i-1];
			while (i--) {
				if (sp->active)
					(*sDef[sp->type].eraseFunc)(sp);
				--sp;
			}

			// Erase Stars
			EraseStars();

			EraseRadar();

			if (!gClientMode) {
				// Move Sprites - using Sprite's moveFunction
				// 4/414/95 Moved Up Here...
				//
				n = gMaxSprite;
				for (i = 0,sp=sTable; i < n; ++i,++sp) {
					// Update Sprite Position
					if (sp->active)
						(*sDef[sp->type].moveFunc)(sp);
				}
				TrackScore();

				GenNewEvents();
	
				// Check Keyboard - respond to keys
				CheckKeys();
	
				if (gConnectionType)
					ServerIdle();
			}
			else {
				CheckKeys();
				if (gConnectionType)
					ClientIdle();
			}


			// Draw Stars
			FlickerStars();

			StatusUpdate();

			// Draw Sprites in Offscreen GWorld
			//
			n = gMaxSprite;
			for (i = 0,sp=sTable; i < n; ++i,++sp) {
				// Render Sprite in New Position
				if (sp->active)
					(*sDef[sp->type].drawFunc)(sp);
			}

			UpdateStars();		// Update Stars Onscreen
	
			// Render Sprites onto Onscreen video
			//
			n = gMaxSprite;
			for (i = 0,sp=sTable; i < n; ++i,++sp) {
				// Render Sprite in New Position, if it needs updating
				//
				if (sp->update) {
					register Point	p1,p2;

					sp->update = false;

					p1 = sp->oldPos;
					p2 = sp->pos;
					p1.h += (gUnivWidth - gOldOffset.h); p1.h %= gUnivWidth;	// 3/7/95 scrolling playfield
					p1.v += (gUnivHeight - gOldOffset.v); p1.v %= gUnivHeight;
					p2.h += (gUnivWidth - gOffsetPt.h); p2.h %= gUnivWidth;
					p2.v += (gUnivHeight - gOffsetPt.v); p2.v %= gUnivHeight;
					// If distance between old and new positions is large
					// than sprite has crossed boundaries from one side of screen
					// to the other, in this case the part of the sprite that was
					// erased is on the other side of the screen and needs to
					// be drawn separately
					//
					if (abs(p2.h - p1.h) > 320 ||
						abs(p2.v - p1.v) > 240)
					{
						// Object has crossed boundaries, draw two separate rects
						r.left = p1.h;
						r.top =  p1.v;
						r.right = p1.h+sp->width;
						r.bottom = p1.v+sp->width;
						// Clip
						if (r.left < SafeWidth)
							r.left = SafeWidth;
						if (r.top < SafeHeight+12)
							r.top = SafeHeight+12;
						if (r.right >= gPlayRect.right)
							r.right = gPlayRect.right-1;
						if (r.bottom >= gPlayRect.bottom)
							r.bottom = gPlayRect.bottom-1;
						if (r.right > r.left && r.bottom > r.top)
							MyCopyRect(&r);
						r.left = p2.h;
						r.top =  p2.v;
						r.right = p2.h+sp->width;
						r.bottom = p2.v+sp->width;
						// Clip
						if (r.left < SafeWidth)
							r.left = SafeWidth;
						if (r.top < SafeHeight+12)
							r.top = SafeHeight+12;
						if (r.right >= gPlayRect.right)
							r.right = gPlayRect.right-1;
						if (r.bottom >= gPlayRect.bottom)
							r.bottom = gPlayRect.bottom-1;
						if (r.right > r.left && r.bottom > r.top)
							MyCopyRect(&r);
					}
					// Otherwise, the part that was erased is nearby
					// we merge those two rects into one rect and
					// copy the whole area in one pass
					//
					else {
						// Merge rects into composite rect
						r.left = (p1.h < p2.h)? p1.h : p2.h;
						r.top = (p1.v < p2.v)? p1.v : p2.v;
						r.right = (p1.h > p2.h)? p1.h : p2.h;
						r.right += sp->width;
						r.bottom = (p1.v > p2.v)? p1.v : p2.v;
						r.bottom += sp->width;
						// Clip
						if (r.left < SafeWidth)
							r.left = SafeWidth;
						if (r.top < SafeHeight+12)
							r.top = SafeHeight+12;
						if (r.right >= gPlayRect.right)
							r.right = gPlayRect.right-1;
						if (r.bottom >= gPlayRect.bottom)
							r.bottom = gPlayRect.bottom-1;
						if (r.right > r.left && r.bottom > r.top)
							MyCopyRect(&r);
					}
					sp->oldPos = sp->pos;
				}
			}
			gOldOffset = gOffsetPt;
			UpdateRadar();
#if DEBUGGING
			if (gDebug) {
				while (Button())
					;
				while (!Button())
					;
			}
#endif
		}
#if DEBUGGING
		// This debugging function is used with the profiler to determine
		// how much idle time I am getting on a slow machine
		// this was used to determine how fast I could make the game clock
		// it is currently set at 20 frames a second
		else {
			IdleTest();
		}
#endif
	// Keep looping as long as game is in play
	// If game is over, keep looping while there are visible sparks from the
	// last explosion
	} while (gGameState == GS_Play || (gGameState == GS_GameOver && gSparkCnt));
#if __option(profile)			// 6/15 Optional profiling support
	_profile = 0;
#endif
	FlushEvents(everyEvent,0);
}

void InitStatusDisplay(void)
{
	gLastScore = -1;
	gLastDrawnScore = -1;
	gLastRemainingShips = -1;
	gLastLevel = -1;
	gLastScoreMultiply = -1;
	gLastShieldPower = -1;
}

// Two different sets of X coordinates are used for the status bar
// elements depending on whether I'm on a 12 inch monitor or not

enum {XScore, XLevel, XPots, XMultiply, XPause, XShields};

static short	xP[2][6] = {{8,90,   170,   254,   320,   488},
							{8,90-32,170-32,254-32,320-32,488-128}};

// Draw the game status bar on top of the screen
//
void StatusUpdate()
{
	static char		scoreText[32];
	Point			loc = {64,40};

	if (gLastDrawnScore == gGameScore &&
		gLastShieldPower == gShieldPower &&
		gLastLevel == gGameLevel &&
		gLastRemainingShips == gLastRemainingShips &&
		gLastScoreMultiply == gScoreMultiply)
		return;

	SetPort(gWindow);

	gLastDrawnScore = gGameScore;
	StatusPrintf(xP[g12InchMode][XScore],12,"%-10ld",gGameScore);

	gLastLevel = gGameLevel;
	StatusPrintf(xP[g12InchMode][XLevel],12,"Level: %d", gGameLevel);

	gLastRemainingShips = gRemainingShips;
	StatusPrintf(xP[g12InchMode][XPots],12,"Pots: %-3d", gRemainingShips);

	gLastScoreMultiply = gScoreMultiply;
	if (gScoreMultiply > 1)
		StatusPrintf(xP[g12InchMode][XMultiply], 12,"x%ld     ", gScoreMultiply);
	else
		StatusPrintf(xP[g12InchMode][XMultiply], 12,"        ", gScoreMultiply);

	{
		RGBColor	fc;

		gLastShieldPower = gShieldPower;
		MoveTo(gVidOffset.h+xP[g12InchMode][XShields],gVidOffset.v+8);

		if (gShieldPower) {
			fc.red = 0x0000;	fc.green = fc.blue = 0x8080;
			RGBForeColor(&fc);
			Line(gShieldPower,0);
		}

		if (MaxShieldPower - gShieldPower) {
			fc.red = 0xFFFF;	fc.green = fc.blue = 0x0000;
			RGBForeColor(&fc);
			Line(MaxShieldPower-gShieldPower,0);
		}

		fc.red = fc.green = fc.blue = 0x0000;
		RGBForeColor(&fc);
	}
}

void TrackScore()
{
	if (gLastScore != gGameScore) {
		if (gGameScore / NewLevelIncrement != gLastScore / NewLevelIncrement) {
			EndLevel();
			InitLevel();
		}
		if ((gGameScore >= 10000L && gLastScore < 10000) ||
			gGameScore/FreeShipPoints != gLastScore/FreeShipPoints) {
			++gRemainingShips;
			PlaySound(S_ExtraShip, 4);
		}
		gLastScore = gGameScore;
	}
}

// Keep track of game score
//
void AddScore(short amt)
{
	gGameScore += amt * gScoreMultiply;
}

// Print part of top score display
//
void PrintfXY(short x, short y, char *template, ...)
{
	char 			tbuf[128];
	va_list 		args;
	register		short cw;

	va_start(args,template);
	vsprintf(tbuf,template,args);
	va_end(args);

	CtoPstr(tbuf);

	cw = StringWidth((StringPtr) tbuf);
	MoveTo(x,y);
	DrawString((StringPtr) tbuf);
}

// Move to the right position to center a string
//
void CenterString(StringPtr str)
{
	MoveTo(((gOffscreenRect.right - gOffscreenRect.left)-StringWidth(str))/2,
			(gOffscreenRect.bottom - gOffscreenRect.top)/2);
}

// Print part of status bar display
//
void StatusPrintf(short x, short y, char *template, ...)
{
	char 			tbuf[128];
	va_list 		args;
	register		short cw;

	va_start(args,template);
	vsprintf(tbuf,template,args);
	va_end(args);

	CtoPstr(tbuf);

	cw = StringWidth((StringPtr) tbuf);
	MoveTo(x+gVidOffset.h,y+gVidOffset.v);
	DrawString((StringPtr) tbuf);
}

// Keyboard handling
// for speed, we use GetKeys function which polls entire keyboard
// bypassing normal WaitNextEvent handlng
//
// DeBounce is used to prevent rapid firing
// in "Uzi" mode, debounce is less to allow rapid firing
//

void PauseMode()
{
	KeyMap	km;

	// !!! Pausing will screw things up if networked...

	do {
		GetKeys(km);
	} while (((char *) km)[4] & 0x08);	// Wait for Key Up

	// Print Pause Message
	StatusPrintf(xP[g12InchMode][XPause],12,"PAUSE");
	gGameState = GS_Pause;
	// FlushEvents(everyEvent,0);
	
	do {
		GetKeys(km);
	} while (!(((char *) km)[4] & 0x08));	// Wait for Key Down
	do {
		GetKeys(km);
	} while (((char *) km)[4] & 0x08);	// Wait for Key Up

	gGameState = GS_Play;
	// Print UnPause Message
	StatusPrintf(xP[g12InchMode][XPause],12,"             ");
}

void CheckKeys(void)
{
	static KeyMap	km;

	GetKeys(km);


	// Non Ship Related Stuff
	//
	if ((((char *) km)[1] & 0x10) ||		// 0x0C Q
		(((char *) km)[6] & 0x20)) {		// 0x35 ESC
		gGameState = GS_GameOver;
#if DEBUGGING
		gDebug = 0;
#endif
		return;
	}

#if DEBUGGING
	if (((char *) km)[5] & 0x04) {			// 0x2A \		trigger random event
		switch (MyRandom(1)) {
		case 0:	NewToaster();				break;
		case 1:	NewSaucer(); 				break;
		case 2: NewCroissant();				break;
		case 3: NewCube();					break;
		}
	}
	if (((char *) km)[6] & 0x04) {			// 0x32 ~		trigger random event
		switch (MyRandom(1)) {
		case 0: NewMixer();					break;
		}
	}
#endif

	if (((char *) km)[4] & 0x08)			// 0x23 P
		PauseMode();

	// Ship Control Keys
	//
	gKeyFlags = 0;

	if (((char *) km)[gPrefs.thrustKeyByte] & gPrefs.thrustKeyBit)
		gKeyFlags |= KF_Thrust;

	if (((char *) km)[gPrefs.fireKeyByte] & gPrefs.fireKeyBit)
		gKeyFlags |= KF_Fire;

	if (((char *) km)[gPrefs.leftKeyByte] & gPrefs.leftKeyBit)
		gKeyFlags |= KF_Left;

	if (((char *) km)[gPrefs.rightKeyByte] & gPrefs.rightKeyBit)
		gKeyFlags |= KF_Right;

	if (((char *) km)[gPrefs.shieldKeyByte] & gPrefs.shieldKeyBit)
		gKeyFlags |= KF_Shield;
}

#if DEBUGGING
void IdleTest()
{
	short	x;
	x = 0;
}
#endif


short			oldMBarHeight;
Boolean		gHiddenBar;

void HideMenuBar(WindowPtr theWin)
{
	if (!gHiddenBar) {
		oldMBarHeight = GetMBarHeight();
		oldMBarHeight = LMGetMBarHeight();
		LMSetMBarHeight(0);
		RectRgn(theWin->visRgn,&qd.screenBits.bounds);
		gHiddenBar = true;
	}
}

void RestoreMenuBar()
{
	if (gHiddenBar) {
		LMSetMBarHeight(oldMBarHeight);
		DrawMenuBar();
		gHiddenBar = false;
	}
}


void RefixWindow()
{
	RectRgn(gWindow->visRgn,&qd.screenBits.bounds);
}