// Radar.c

#include "CToast.h"

GWorldPtr		gRadarScreen;
PixMapHandle	gRadarMap;
Ptr				gRadarMem;
long			gRadarRowBytes;
Rect			gRadarRect;
Point			gRadarOffset,gRadarVidOffset;
Ptr				gRadarVidBase;
Handle			gRadarCircle;

void ToggleRadar();

#define RadarWidth		128
#define UnitsPerRadar	16
#define RadarShift		4
#define RadarUnivWidth	240	// gUnivWidth/16

void DrawRadarCircle();

void InitRadar()
{
	OSErr		oe;
	CGrafPtr	curPort;
	GDHandle	curDevice;
	PicHandle	radarPic;

	gRadarCircle = GetAsset('LSPR', 153);
	MoveHHi(gRadarCircle);
	HLock(gRadarCircle);

	SetRect(&gRadarRect,0,0,RadarWidth,RadarWidth);
	if ((oe = NewGWorld(&gRadarScreen,0,&gRadarRect,NULL,NULL,0)) != noErr) {
		DebugStr("\pOy!");
		return;
	}
	gRadarMap = GetGWorldPixMap(gRadarScreen);
	LockPixels(gRadarMap);
	gRadarMem = (*gRadarMap)->baseAddr;
	gRadarRowBytes = (*gRadarMap)->rowBytes & 0x3FFF;

	GetGWorld(&curPort,&curDevice);

	SetGWorld(gRadarScreen,NULL);
	PaintRect(&gRadarScreen->portRect);
	radarPic = (PicHandle) GetAsset('PICT',RadarMaskPICT);
	DrawPicture(radarPic,&gRadarRect);
	ReleaseAsset((Handle) radarPic);
	SetGWorld(curPort,curDevice);
	gRadarVidOffset.h = gVidOffset.h+gVisPlayWidth-RadarWidth;
	gRadarVidOffset.v = gVidOffset.v+gVisPlayHeight-RadarWidth;
	gRadarVidBase = gVideoMem+ (gRadarVidOffset.v * gVideoRowBytes) + gRadarVidOffset.h;

}


void DrawRadarSprite(SpriteInstance *sp)
{
	Point			loc;
	unsigned char	*vidPtr,*scrnPtr;

	// Find Position in Radar Map

	loc = sp->pos;
	loc.h >>= RadarShift;
	loc.v >>= RadarShift;
	loc.h += (RadarUnivWidth - gRadarOffset.h);	loc.h %= RadarUnivWidth;
	loc.v += (RadarUnivWidth - gRadarOffset.v);	loc.v %= RadarUnivWidth;

	// If not within Radar map, return
	if (!PtInRect(loc,&gRadarRect)) {
		sp->savePixel = 0;
		return;
	}

	// If Radar Map pixel is white return
	scrnPtr = (unsigned char *) (gRadarMem + (loc.v*gRadarRowBytes + loc.h));
	if (*scrnPtr == 0)
		return;

	vidPtr = (unsigned char *) (gRadarVidBase + (loc.v*gVideoRowBytes + loc.h));

	sp->savePixel = *vidPtr;

	*vidPtr = gRadarColor[sp->type];
}

void EraseRadarSprite(SpriteInstance *sp)
{
	Point			loc;
	unsigned char	*vidPtr;

	// Find Position in Radar Map
	if (sp->savePixel == 0)
		return;

	loc = sp->pos;
	loc.h >>= RadarShift;
	loc.v >>= RadarShift;
	loc.h += (RadarUnivWidth - gRadarOffset.h);	loc.h %= RadarUnivWidth;
	loc.v += (RadarUnivWidth - gRadarOffset.v);	loc.v %= RadarUnivWidth;

	// If not within Radar map, return
	if (!PtInRect(loc,&gRadarRect)) {
		sp->savePixel = 0;
		return;
	}

	// If Radar Map pixel is white return
	vidPtr = (unsigned char *) (gRadarVidBase + (loc.v*gVideoRowBytes + loc.h));
	*vidPtr = 0xFF;	// sp->savePixel[0];
	sp->savePixel = 0;

}

void EraseRadar()
{
	short			i;
	SpriteInstance	*sp;
	SInt8			mmode;

	mmode = true32b;
	SwapMMUMode(&mmode);

	for (i = 0,sp = &sTable[gMaxSprite-1]; i < gMaxSprite; ++i,--sp)
		if (sp->active && gRadarColor[sp->type] != Invisible)
			EraseRadarSprite(sp);

	SwapMMUMode(&mmode);

	DrawRadarCircle();
}

void UpdateRadar()
{
	short			i;
	SpriteInstance	*sp;
	SInt8			mmode;

	mmode = true32b;
	SwapMMUMode(&mmode);

	gRadarOffset.h = (gShip->pos.h >> RadarShift) + (RadarUnivWidth - RadarWidth/2);
	gRadarOffset.v = (gShip->pos.v >> RadarShift) + (RadarUnivWidth - RadarWidth/2);
	gRadarOffset.h %= RadarUnivWidth;
	gRadarOffset.v %= RadarUnivWidth;

	for (i = 0,sp = sTable; i < gMaxSprite; ++i,++sp)
		if (sp->active && gRadarColor[sp->type] != Invisible)
			DrawRadarSprite(sp);

	SwapMMUMode(&mmode);
}

void DrawRadarCircle()
{
	unsigned char	*sprPtr,*scrPtr;
	LSprite			*lSpriteHeader;
	short			j,k;
	unsigned char	typ,cnt;
	long			rowBytesOffset;
	Point			loc;

	// return;
	
	loc.h = SafeWidth+gVisPlayWidth-RadarWidth;
	loc.v = SafeHeight+gVisPlayHeight-RadarWidth;

	rowBytesOffset = gScreenRowBytes - RadarWidth;

	lSpriteHeader = (LSprite *) *gRadarCircle;
	sprPtr = (unsigned char *) &lSpriteHeader->data[lSpriteHeader->dataOffset];
	scrPtr = (unsigned char *) (gScreenMem + (loc.v*gScreenRowBytes + loc.h));

	for (j = lSpriteHeader->height; j; j--) {
		k = lSpriteHeader->width;
		sprPtr += sizeof(short);	// Skip Count
		while (k) {
			typ = *sprPtr & LSPR_HiBits;
			cnt = *sprPtr & LSPR_LowBits;
			k -= cnt;
			++sprPtr;
			switch (typ) {
			case LSPR_Mask:
				scrPtr += cnt;
				break;
			case LSPR_Opaque:
				while (cnt--)
					*(scrPtr++) = *(sprPtr++);
				break;
			case LSPR_Mixed: // unsupported
				while (cnt--)
					*(scrPtr++) = *(sprPtr++);
				break;
			}
		}
		scrPtr += rowBytesOffset;
	}
}
