/****************************************************************************
 * Sprites.c
 *
 *		Sprite Maintenence, Logic, Animation
 *
 *		Most of Game Logic is here...
 *
 * 2/28/95
 *		Updated to compile using newer Think C / Metrowerks compiler
 *		Put limit on maximum ship speed.
 ****************************************************************************/
#include "CToast.h"
#include <stdlib.h>
#include <math.h>

#if __option(profile)					// 6/15 Optional profiling support
#include <Profile.h>
#endif


// Sprites Initialization Functions

SpriteInstance *NewSprite()
{
	SpriteInstance *sp;
	if (gMaxSprite == MaxSprites - 1) {
		// DebugStr("\pMax Sprites");
		return NULL;
	}
	sp = &sTable[gMaxSprite];
	sp->active = true;
	if (!gClientMode)
		sp->id = spriteIDCtr++;
	else
		sp->id = 0L;
	++gMaxSprite;
	++gSpriteCnt;
	return sp;
}


void KillSprite(SpriteInstance *sp)
{
	sp->active = 0;
	--gSpriteCnt;
	while (gMaxSprite > 0 && !sTable[gMaxSprite-1].active && !sTable[gMaxSprite-1].update)
		--gMaxSprite;
}

void KillNetSprite(SpriteInstance *sp)
{
	SendKillSprite(sp->id);
	KillSprite(sp);
}

void NewAsteroid(short type)
{
	register short	vecSpeed;
	register SpriteInstance *sp;

	if ((sp = NewSprite()) == NULL)
		return;
	if (MyRandom(50) == 0)
		type = ST_BigJim;
	sp->type = type;
	sp->aniState = MyRandom(sDef[sp->type].nbrIcons);
	sp->pos.h = MyRandom(gUnivWidth);
	sp->pos.v = MyRandom(gUnivHeight);
	sp->oldPos = sp->pos;
	// Pick any angle except 0,4,8,12 (which are straight)
	// and may cause asteroid to be invisible on sides
	sp->angle = MyRandom(MaxAngles-4);
	sp->angle += 1+(sp->angle/7);
	vecSpeed = MyRandom(3)+2+MyRandom(gGameLevel);
	sp->vector.lh = vecTable[sp->angle].lh * vecSpeed;
	sp->vector.lv = vecTable[sp->angle].lv * vecSpeed;
	if (labs(sp->vector.lh) < 0x00010000)
		sp->vector.lh = 0x00010000;
	if (labs(sp->vector.lv) < 0x00010000)
		sp->vector.lv = 0x00010000;
	sp->aniSpeed = 1 + MyRandom(3);
	sp->tickCtr = MyRandom(sp->aniSpeed);
	sp->width = sDef[type].class == SP_Large? LargeIconWidth : IconWidth;
	++gAsteroidCnt;
	SendNewSprite(sp);
}

#if DEBSPRITES
// This puts two special sprites in the lower left hand corner of the screen
// which are used to monitor sprite memory
//
void NewDebugDisplay(void)
{
	register SpriteInstance *sp;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_SpriteCnt;
	sp->state = 0;
	sp->pos.h = 40;
	sp->pos.v = 480+32;
	sp->oldPos = sp->pos;
	sp->width = 2;

	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_MaxSprite;
	sp->state = 0;
	sp->pos.h = 32;
	sp->pos.v = 480+32;
	sp->oldPos = sp->pos;
	sp->width = 2;
}
#endif

void NewSubAsteroid(SpriteInstance *par, short type, short vecOffset)
{
	register short vecSpeed;
	register SpriteInstance *sp;

	if ((sp = NewSprite()) == NULL)
		return;
	sp->pos = par->pos;
	sp->oldPos = par->oldPos;
	switch (type) {
	case ST_SmallCup:
	case ST_Jim:
		sp->aniState = MyRandom(sDef[type].nbrIcons);
		break;
	default:
		sp->aniState = par->aniState;
		if (sp->aniState >= sDef[type].nbrIcons)
			sp->aniState = MyRandom(sDef[type].nbrIcons);
		break;
	}
	sp->lifeSpan = par->lifeSpan;
	sp->angle = par->angle;
	sp->vector = par->vector;

	sp->type = type;
	sp->angle += vecOffset;
	if (sp->angle < 0)
		sp->angle += MaxAngles;
	if (sp->angle >= MaxAngles)
		sp->angle -= MaxAngles;
	vecSpeed = MyRandom(3)+2+MyRandom(gGameLevel);
	sp->vector.lh = vecTable[sp->angle].lh * vecSpeed;
	sp->vector.lv = vecTable[sp->angle].lv * vecSpeed;
	sp->aniSpeed = 1 + MyRandom(2);
	sp->tickCtr = 0;
	sp->width = IconWidth;
	++gAsteroidCnt;
	SendNewSprite(sp);
}

void NewSaucer(void)
{
	register SpriteInstance *sp;
	short	t;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_Saucer;
	if ((t = 2+gGameLevel/2) > 255)
		sp->maxHits = 255;
	else
		sp->maxHits = t;			// Max Hits
	sp->hitsTaken = 0;							// Hits Taken
	sp->aniState = MyRandom(sDef[sp->type].nbrIcons);
	sp->pos.h = MyRandom(gUnivWidth);
	sp->pos.v = MyRandom(gUnivHeight);
	sp->oldPos = sp->pos;
	sp->angle = MyRandom(MaxAngles-4);
	sp->angle += 1+sp->angle/7;
	sp->vector.lh = vecTable[sp->angle].lh;
	sp->vector.lv = vecTable[sp->angle].lv;
	sp->aniSpeed = 3 - (gGameLevel > 1? 1 : 0);
	sp->tickCtr = 0;
	sp->width = IconWidth;
	PlaySound(S_BadGuyDebut, 3);
	SendNewSprite(sp);
}

void NewBeater(short x, short y, short angle)
{
	register SpriteInstance *sp;
	register short t;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_Beater;
	if ((t = 2+gGameLevel/2) > 255)
		sp->maxHits = 255;
	else
		sp->maxHits = t;			// Max Hits
	sp->hitsTaken = 0;							// Hits Taken
	sp->pos.h = x;
	sp->pos.v = y;
	sp->oldPos = sp->pos;
	sp->aniState = angle*3+angle;
	sp->vector.lh = 0;
	sp->vector.lv = 0;
	sp->aniSpeed = 3 - (gGameLevel > 1? 1 : 0);
	sp->tickCtr = 0;
	sp->width = IconWidth;
	PlaySound(S_BadGuyDebut, 3);
	SendNewSprite(sp);
}

void NewMixer()
{
	register SpriteInstance *sp;
	register short t;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_Mixer;
	if ((t = 2+gGameLevel/2) > 255)
		sp->maxHits = 255;
	else
		sp->maxHits = t;			// Max Hits
	sp->hitsTaken = 0;							// Hits Taken
	sp->aniState = 0;
	sp->pos.h = MyRandom(gUnivWidth);
	sp->pos.v = MyRandom(gUnivHeight);
	sp->oldPos = sp->pos;
	sp->angle = MyRandom(MaxAngles-4);
	sp->angle += 1+sp->angle/7;
	sp->vector.lh = vecTable[sp->angle].lh;
	sp->vector.lv = vecTable[sp->angle].lv;
	sp->aniSpeed = 10 - (gGameLevel > 1? 1 : 0);
	sp->tickCtr = 0;
	sp->width = LargeIconWidth;
	PlaySound(S_BadGuyDebut, 3);
	++gMixerCount;
	SendNewSprite(sp);
}


void NewCoffeePot(void)
{
	register SpriteInstance *sp;
	register short t;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_CoffeePot;
	if ((t = 2+gGameLevel/2) > 255)
		sp->maxHits = 255;
	else
		sp->maxHits = t;			// Max Hits
	sp->hitsTaken = 0;							// Hits Taken
	sp->aniState = MyRandom(sDef[sp->type].nbrIcons);
	sp->pos.h = MyRandom(gUnivWidth);
	sp->pos.v = MyRandom(gUnivHeight);
	sp->oldPos = sp->pos;
	sp->angle = MyRandom(MaxAngles-4);
	sp->angle += 1+sp->angle/7;
	sp->vector.lh = vecTable[sp->angle].lh;
	sp->vector.lv = vecTable[sp->angle].lv;
	sp->aniSpeed = 3 - (gGameLevel > 1? 1 : 0);
	sp->tickCtr = 0;
	sp->width = LargeIconWidth;
	PlaySound(S_KevinsPot, 3);
	SendNewSprite(sp);
}

void NewCroissant(void)
{
	register SpriteInstance *sp;
	register short t;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_Croissant;
	if ((t = 3+gGameLevel/2) > 255)
		sp->maxHits = 255;
	else
		sp->maxHits = t;			// Max Hits
	sp->hitsTaken = 0;							// Hits Taken
	sp->aniState = MyRandom(sDef[sp->type].nbrIcons);
	sp->pos.h = MyRandom(gUnivWidth);
	sp->pos.v = MyRandom(gUnivHeight);
	sp->oldPos = sp->pos;
	sp->angle = MyRandom(MaxAngles-4);
	sp->angle += 1+sp->angle/7;
	sp->vector.lh = vecTable[sp->angle].lh;
	sp->vector.lv = vecTable[sp->angle].lv;
	sp->aniSpeed = 3 - (gGameLevel > 1? 1 : 0);
	sp->tickCtr = 0;
	sp->width = LargeIconWidth;
	sp->lifeSpan = CroissantLifeSpan;
	PlaySound(S_BadGuyDebut, 3);
	SendNewSprite(sp);
}

void NewCube(void)
{
	register SpriteInstance *sp;
	register short t;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_Cube;
	if ((t = 4+gGameLevel/2) > 255)
		sp->maxHits = 255;
	else
		sp->maxHits = t;			// Max Hits
	sp->hitsTaken = 0;							// Hits Taken
	sp->aniState = MyRandom(sDef[sp->type].nbrIcons);
	sp->pos.h = MyRandom(gUnivWidth);
	sp->pos.v = MyRandom(gUnivHeight);
	sp->oldPos = sp->pos;
	sp->angle = MyRandom(MaxAngles-4);
	sp->angle += 1+sp->angle/7;
	sp->vector.lh = vecTable[sp->angle].lh;
	sp->vector.lv = vecTable[sp->angle].lv;
	sp->aniSpeed = 3 - (gGameLevel > 1 ? 1 : 0);
	sp->tickCtr = 0;
	sp->width = IconWidth;
	sp->lifeSpan = CubeLifeSpan;
	PlaySound(S_BadGuyDebut, 3);
	SendNewSprite(sp);
}

void NewToaster(void)
{
	register SpriteInstance *sp;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_Toaster;
	sp->lifeSpan = ToasterLifeSpan;
	sp->aniState = MyRandom(sDef[sp->type].nbrIcons);
	sp->pos.h = MyRandom(gUnivWidth);
	sp->pos.v = MyRandom(gUnivHeight);
	sp->oldPos = sp->pos;
	sp->angle = 16;	// Straight Down
	sp->vector.lh = vecTable[sp->angle].lh << 4;
	sp->vector.lv = vecTable[sp->angle].lv << 4;
	sp->aniSpeed = 3;
	sp->tickCtr = 0;
	sp->width = LargeIconWidth;
	sp->stateTimer = 0;
	sp->state = 0;
	// PlaySound(S_BadGuyDebut, 3);
	++gToasterCount;
	SendNewSprite(sp);
}

void LaunchNookandCranny(SpriteInstance *par)
{
	register SpriteInstance *sp,*sp2;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_EMuffin;
	sp->aniState = 0;
	sp->pos.h = par->pos.h+2;
	sp->pos.h += gUnivWidth;	sp->pos.v %= gUnivWidth;
	sp->pos.v = par->pos.v+6;
	sp->pos.v += gUnivHeight;	sp->pos.v %= gUnivHeight;
	sp->oldPos = sp->pos;
	sp->angle = 0;	// Straight Right?
	sp->vector.lh = vecTable[sp->angle].lh << 3;
	sp->vector.lv = vecTable[sp->angle].lv << 3;
	sp->aniSpeed = 2;
	sp->tickCtr = 0;
	sp->width = IconWidth;
	sp->state = 0;
	SendNewSprite(sp);

	if ((sp2 = NewSprite()) == NULL)
		return;
	sp2->type = ST_EMuffin;
	sp2->aniState = 0;
	sp2->pos.h = par->pos.h+11;
	sp->pos.h += gUnivWidth;	sp->pos.v %= gUnivWidth;
	sp2->pos.v = par->pos.v+9;
	sp->pos.v += gUnivHeight;	sp->pos.v %= gUnivHeight;
	sp2->oldPos = sp->pos;
	sp2->angle = sp->angle;
	sp2->vector = sp->vector;
	sp2->aniSpeed = 2;
	sp2->tickCtr = 0;
	sp2->width = IconWidth;
	sp2->state = 1;
	SendNewSprite(sp2);
}



void NewShipTimer(register SpriteInstance *sp)
{
	if (sp->tickCtr++ > 60) {

		// Reset scroll origin to ship base...
		gOffsetPt.v = 0;
		gOffsetPt.h = 0;

		sp->tickCtr = 0;
		sp->type = ST_Teapot;
		PlaySound(S_ShipDebut, 3);
		gShipMode |= SM_AutoShield;
		gShieldPower = MaxShieldPower;
	}
}

void NewShip(void)
{
	register SpriteInstance *sp;
	if (gRemainingShips == 0) {
		gGameState = GS_GameOver;
		return;
	}
	if ((sp = NewSprite()) == NULL)
		return;
	gShip = sp;
	sp->type = ST_TeapotNew;
	sp->state = 0;
	sp->pos.h = gCenterP.h-24;
	sp->pos.v = gCenterP.v-24;
	sp->oldPos = sp->pos;
	sp->angle = 0;
	sp->aniState = 0;
	sp->vector.lh = 0L;
	sp->vector.lv = 0L;
	sp->aniSpeed = 0;
	sp->tickCtr = 0;
	sp->width = LargeIconWidth;
	--gRemainingShips;
	gShipMode = 0;
	SendNewSprite(sp);
}

void NewYummy(void)
{
	register SpriteInstance *sp;
	short	yummyType;
	yummyType = MyRandom(NbrYummies);
	if (yummyType == Y_Creamer) {
		NewCreamer();
		return;
	}
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_Yummies;
	sp->aniState = yummyType;
	sp->state = sp->aniState;
	sp->pos.h = MyRandom(gUnivWidth);
	sp->pos.v = MyRandom(gUnivHeight);
	sp->vector.lh = 0;
	sp->vector.lv = 0;
	sp->oldPos = sp->pos;
	sp->lifeSpan = YummyLifeSpan;
	sp->tickCtr = 0;
	sp->width = IconWidth;
	// StandardSpriteDraw(sp);
	++gYummyCnt;
	SendNewSprite(sp);
}

void NewCreamer(void)
{
	register SpriteInstance *sp;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_Creamer;
	sp->state = 6;			// Yummy Subtype
	sp->aniState = MyRandom(sDef[sp->type].nbrIcons);
	sp->pos.h = MyRandom(gUnivWidth);
	sp->pos.v = MyRandom(gUnivHeight);
	sp->lifeSpan = CreamerLifeSpan;
	sp->oldPos = sp->pos;
	sp->aniSpeed = 1;
	sp->tickCtr = 0;
	sp->width = LargeIconWidth;
	++gYummyCnt;
	SendNewSprite(sp);
}

void NewQuake(void)
{
	register SpriteInstance *sp;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_Quake;
	sp->lifeSpan = QuakeLifeSpan;
	sp->pos.h = sp->pos.v = 0;
	sp->vector.lh = sp->vector.lv = 0;
	SendNewSprite(sp);
}

void NewExplosion(SpriteInstance *parent)
{
	register SpriteInstance *sp;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_Explosion;
	sp->aniState = 0;
	sp->pos = parent->pos;
	sp->oldPos = sp->pos;
	sp->vector = parent->vector;
	sp->aniSpeed = 2;
	sp->tickCtr = 0;
	sp->width = LargeIconWidth;
	SendNewSprite(sp);
}


void LaunchBullet(short type, short x, short y, long vx, long vy, short lifeSpan, short bFlag)
{
	register SpriteInstance *sp;
	if (type == ST_Bullet && (gShipMode & SM_CreamBullet))
		type = ST_CreamBullet;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = type;
	sp->targetFlags = bFlag;
	sp->state = 0;
	sp->pos.h = x;
	sp->pos.v = y;
	if (type == ST_CreamBullet) {
		sp->pos.h -= 8;
		sp->pos.v -= 8;
	}
	sp->oldPos = sp->pos;
	sp->vector.lh = vx;
	sp->vector.lv = vy;
	sp->tickCtr = 0;
	sp->lifeSpan = lifeSpan;
	sp->width = (type == ST_CreamBullet? 16 : 2);
	sp->aniState = 0;
	SendNewSprite(sp);
}

void LaunchSpark(short x, short y, long vx, long vy, short lifeSpan, short color)
{
	register SpriteInstance *sp;
	if ((sp = NewSprite()) == NULL)
		return;
	sp->type = ST_Spark;
	sp->state = color;		// Spark Color
	sp->pos.h = x;
	sp->pos.v = y;
	sp->oldPos = sp->pos;
	sp->vector.lh = vx;
	sp->vector.lv = vy;
	sp->tickCtr = 0;
	sp->lifeSpan = lifeSpan;
	sp->width = 2;
	++gSparkCnt;
}

// Specialized Sprites - Kill Functions
void ConsumeYummy(register SpriteInstance *sp)
{
	switch (sp->state) {
	case Y_Mult2:		gScoreMultiply += (gScoreMultiply == 1? 1 : 2);		break;
	case Y_Mult3:		gScoreMultiply += (gScoreMultiply == 1? 2 : 3);		break;
	case Y_Mult5:		gScoreMultiply += (gScoreMultiply == 1? 4 : 5);		break;
	case Y_Triple:		gShipMode |= SM_Triple;								break;
	case Y_Uzi:			gShipMode |= SM_Uzi;								break;
	case Y_Shield:		gShieldPower = MaxShieldPower;						break;
	case Y_Creamer:		gShipMode |= SM_CreamBullet;						break;
	}
	KillNetSprite(sp);
	PlaySound(S_YummyConsume, 3);
	--gYummyCnt;
}


// Explosions only happen to Net Sprites!!

void ExplodeSprite(register SpriteInstance *sp, register short lifeSpan)
{
	register short	r;

	// deal with it later...
	unsigned char  *sprPtr;
	LSprite			*spH;
	short			j,k;
	unsigned char	typ,cnt,n,speedCtr;
	static unsigned char	speedShift[5] = {1,1,2,1,1};

	SendExplodeSprite(sp->id,lifeSpan);
											 
	spH = (LSprite *) sDef[sp->type].colorMaps;
	sprPtr = &spH->data[spH->frameOffset[sp->aniState]];
	n = 0;
	r = 0;
	speedCtr = 0;
	for (j = 0; j < sp->width; ++j) {
		sprPtr += sizeof(short);
		k = 0;
		while (k < sp->width) {
			typ = *sprPtr & LSPR_HiBits;
			cnt = *sprPtr & LSPR_LowBits;
			++sprPtr;
			switch (typ) {
			case LSPR_Mask:
				k += cnt;
				break;
			case LSPR_Opaque:
				while (cnt--) {
					++n;
					if (n >= 16) {
						n = 0;
						LaunchSpark(sp->pos.h+k,sp->pos.v+j,
									 sp->vector.lh + (vecTable[r].lh << speedShift[speedCtr]),
									 sp->vector.lv + (vecTable[r].lv << speedShift[speedCtr]),
									 lifeSpan,
									 *sprPtr);
						++r;
						r &= (MaxAngles-1);
						++speedCtr;
						if (speedCtr >= 5)
							speedCtr = 0;
					}
					++k;
					++sprPtr;
				}
				break;
			}
		}
	}
	KillSprite(sp);
	if (lifeSpan > 10)
		PlaySound(S_ShipExplodes, 3);
	else
		PlaySound(S_Explosion, 2);
}


// Sprite Move Functions

void StandardSpriteMove(register SpriteInstance *sp)
{
	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	// Update Animation Frame
	sp->tickCtr++;
	if (sp->tickCtr >= sp->aniSpeed) {
		sp->tickCtr = 0;
		sp->aniState++;
		if (sp->aniState >= sDef[sp->type].nbrIcons)
			sp->aniState = 0;
	}
}

#if DEBSPRITES

void MaxSpriteMove(register SpriteInstance *sp)
{
	sp->pos.h = 32;
	sp->pos.v = 480+32-gMaxSprite;
}

void SpriteCntMove(register SpriteInstance *sp)
{
	sp->pos.h = 40;
	sp->pos.v = 480+32-gSpriteCnt;
}

#endif

void CoffeePotMove(register SpriteInstance *sp)
{
	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	// Update Animation Frame
	sp->tickCtr++;
	if (sp->tickCtr >= sp->aniSpeed) {
		sp->tickCtr = 0;
		sp->aniState++;
		if (sp->aniState >= sDef[sp->type].nbrIcons) {
			LongPoint	newVec;
			Point		nVec;
			sp->aniState = 0;
			if (gShip && gShip->type == ST_Teapot || gShip->type == ST_TeapotT) {
				double scale;
				register short	slowness;
				Point	shipPos;
				shipPos = gShip->pos;
				if (shipPos.h - sp->pos.h > gUnivWidth/2)
					shipPos.h -= gUnivWidth;
				if (sp->pos.h - shipPos.h > gUnivWidth/2)
					shipPos.h += gUnivWidth;
				if (shipPos.v - sp->pos.v > gUnivHeight/2)
					shipPos.v -= gUnivHeight;
				if (sp->pos.v - shipPos.v > gUnivHeight/2)
					shipPos.v += gUnivHeight;
				nVec.h = shipPos.h+8 - sp->pos.h;
				nVec.v = shipPos.v+8 - sp->pos.v;
				newVec.lh = nVec.h;
				newVec.lv = nVec.v;
				scale = 1.0 / sqrt(	(double) newVec.lh*newVec.lh +
						newVec.lv*newVec.lv);
				newVec.lh *= scale * 65536.0;
				newVec.lv *= scale * 65536.0;

				if (gGameLevel && MyRandom(3 - (gGameLevel > 1? 1 : 0)) == 0) {
					newVec.lh <<= 4;
					newVec.lv <<= 4;
				}
				else {
					newVec.lh <<= 3;
					newVec.lv <<= 3;
				}
				if (abs(nVec.h) < gVisPlayWidth &&
					abs(nVec.v) < gVisPlayHeight) {
					slowness = 1 + MyRandom(2);
					sp->vector.lh = newVec.lh >> slowness;
					sp->vector.lv = newVec.lv >> slowness;
					LaunchBullet(ST_Photon,
								 sp->pos.h+24,sp->pos.v+4,
								 newVec.lh,newVec.lv,
								 PhotonLifeSpan,
								 BF_BadBulletFlag);
					PlaySound(S_EnemyFires, 2);
				}
				else
					sp->vector = newVec;
			}
		}
		SendSpriteStateChange(sp);
	}
}

void SaucerMove(register SpriteInstance *sp)
{
	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	// Update Animation Frame
	sp->tickCtr++;
	if (sp->tickCtr >= sp->aniSpeed) {
		sp->tickCtr = 0;
		sp->aniState++;
		if (sp->aniState >= sDef[sp->type].nbrIcons) {
			LongPoint	newVec;
			Point		nVec;
			sp->aniState = 0;
			if (gShip && gShip->type == ST_Teapot || gShip->type == ST_TeapotT) {
				double scale;
				register short	slowness;
				Point	shipPos;
				shipPos = gShip->pos;
				if (shipPos.h - sp->pos.h > gUnivWidth/2)
					shipPos.h -= gUnivWidth;
				if (sp->pos.h - shipPos.h > gUnivWidth/2)
					shipPos.h += gUnivWidth;
				if (shipPos.v - sp->pos.v > gUnivHeight/2)
					shipPos.v -= gUnivHeight;
				if (sp->pos.v - shipPos.v > gUnivHeight/2)
					shipPos.v += gUnivHeight;
				nVec.h = shipPos.h+8 - sp->pos.h;
				nVec.v = shipPos.v+8 - sp->pos.v;
				newVec.lh = nVec.h;
				newVec.lv = nVec.v;
				scale = 1.0 / sqrt(	(double) newVec.lh*newVec.lh +
						newVec.lv*newVec.lv);
				newVec.lh *= scale * 65536.0;
				newVec.lv *= scale * 65536.0;

				if (gGameLevel && MyRandom(3 - (gGameLevel > 1? 1 : 0)) == 0) {
					newVec.lh <<= 4;
					newVec.lv <<= 4;
				}
				else {
					newVec.lh <<= 3;
					newVec.lv <<= 3;
				}
				if (abs(nVec.h) < gVisPlayWidth &&
					abs(nVec.v) < gVisPlayHeight) {
					slowness = 1 + MyRandom(2);
					sp->vector.lh = newVec.lh >> slowness;
					sp->vector.lv = newVec.lv >> slowness;
					LaunchBullet(ST_Photon,
								 sp->pos.h+16,sp->pos.v+4,
								 newVec.lh,newVec.lv,
								 PhotonLifeSpan,
								 BF_BadBulletFlag);
					PlaySound(S_EnemyFires, 2);
				}
				else
					sp->vector = newVec;
			}
		}
		SendSpriteStateChange(sp);
	}
}

void CroissantMove(register SpriteInstance *sp)
{
	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	if (--sp->lifeSpan == 0) {
		Point	shipPos,dVec;
		shipPos = gShip->pos;
		if (shipPos.h - sp->pos.h > gUnivWidth/2)
			shipPos.h -= gUnivWidth;
		if (sp->pos.h - shipPos.h > gUnivWidth/2)
			shipPos.h += gUnivWidth;
		if (shipPos.v - sp->pos.v > gUnivHeight/2)
			shipPos.v -= gUnivHeight;
		if (sp->pos.v - shipPos.v > gUnivHeight/2)
			shipPos.v += gUnivHeight;
		dVec.h = shipPos.h+8 - sp->pos.h;
		dVec.v = shipPos.v+8 - sp->pos.v;			
		if (abs(dVec.h) < gVisPlayWidth &&
			abs(dVec.v) < gVisPlayHeight) {
			NewExplosion(sp);
			ExplodeSprite(sp,LargeSparkLifeSpan);
			NewQuake();
		}
		else
			KillNetSprite(sp);
		return;
	}
	// Update Animation Frame
	sp->tickCtr++;
	if (sp->tickCtr >= sp->aniSpeed) {
		sp->tickCtr = 0;
		if (++sp->aniState >= sDef[sp->type].nbrIcons)
			sp->aniState = 0;
		if ((sp->aniState & 1)) {
			LongPoint	bulletVec;
			if (gShip && gShip->type == ST_Teapot || gShip->type == ST_TeapotT) {
				double scale;
				Point	shipPos;
				shipPos = gShip->pos;
				if (shipPos.h - sp->pos.h > gUnivWidth/2)
					shipPos.h -= gUnivWidth;
				if (sp->pos.h - shipPos.h > gUnivWidth/2)
					shipPos.h += gUnivWidth;
				if (shipPos.v - sp->pos.v > gUnivHeight/2)
					shipPos.v -= gUnivHeight;
				if (sp->pos.v - shipPos.v > gUnivHeight/2)
					shipPos.v += gUnivHeight;
				bulletVec.lh = shipPos.h+8 - sp->pos.h;
				bulletVec.lv = shipPos.v+8 - sp->pos.v;			
				scale = 1.0 / sqrt(	(double) bulletVec.lh*bulletVec.lh +
						bulletVec.lv*bulletVec.lv);
				bulletVec.lh *= scale * 65536.0;
				bulletVec.lv *= scale * 65536.0;
				bulletVec.lh <<= 3;	// was 2
				bulletVec.lv <<= 3;	// was 2
				sp->vector = bulletVec;
			}
		}
		SendSpriteStateChange(sp);
	}
}

short GetAngle(LongPoint	*vec)
{
	if (vec->lh > 0) {		// Right Half
		if (vec->lv > 0) {		// Bot Right
			if (vec->lh > vec->lv)
				return 6;
			else
				return 7;
		}
		else {					// Top Right
			if (vec->lh > -vec->lv)
				return 5;
			else
				return 4;
		}
	}
	else {					// Left Half
		if (vec->lv > 0) {		// Bot Left
			if (-vec->lh > vec->lv)
				return 1;
			else
				return 0;
		}
		else {					// Top Left
			if (-vec->lh > -vec->lv)
				return 2;
			else
				return 3;
		}
	}
}

void BeaterMove(register SpriteInstance *sp)
{
	short	angle,frame,tangle;
	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	// Update Animation Frame
	sp->tickCtr++;
	if (sp->tickCtr >= sp->aniSpeed) {
		LongPoint	bulletVec;
		sp->tickCtr = 0;
		if (gShip && gShip->type == ST_Teapot || gShip->type == ST_TeapotT) {
			double scale;
			Point	shipPos;
			shipPos = gShip->pos;
			if (shipPos.h - sp->pos.h > gUnivWidth/2)
				shipPos.h -= gUnivWidth;
			if (sp->pos.h - shipPos.h > gUnivWidth/2)
				shipPos.h += gUnivWidth;
			if (shipPos.v - sp->pos.v > gUnivHeight/2)
				shipPos.v -= gUnivHeight;
			if (sp->pos.v - shipPos.v > gUnivHeight/2)
				shipPos.v += gUnivHeight;
			bulletVec.lh = shipPos.h+8 - sp->pos.h;
			bulletVec.lv = shipPos.v+8 - sp->pos.v;			
			scale = 1.0 / sqrt(	(double) bulletVec.lh*bulletVec.lh +
					bulletVec.lv*bulletVec.lv);
			bulletVec.lh *= scale * 65536.0;
			bulletVec.lv *= scale * 65536.0;
			bulletVec.lh <<= 3;	// was 2
			bulletVec.lv <<= 3;	// was 2
			sp->vector = bulletVec;
		}
	}
	angle = sp->aniState / 3;
	frame = sp->aniState % 3;
	tangle = GetAngle(&sp->vector);
	if (tangle > angle)
		++angle;
	else if (tangle < angle)
		--angle;
	frame = (frame + 1) % 3;
	sp->aniState = angle * 3 + frame;
	SendSpriteStateChange(sp);
}

void MixerMove(register SpriteInstance *sp)
{
	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	// Update Animation Frame
	sp->tickCtr++;
	if (sp->tickCtr >= sp->aniSpeed) {
		LongPoint	bulletVec;
		sp->tickCtr = 0;
		if (gShip && gShip->type == ST_Teapot || gShip->type == ST_TeapotT) {
			double scale;
			Point	shipPos;
			shipPos = gShip->pos;
			if (shipPos.h - sp->pos.h > gUnivWidth/2)
				shipPos.h -= gUnivWidth;
			if (sp->pos.h - shipPos.h > gUnivWidth/2)
				shipPos.h += gUnivWidth;
			if (shipPos.v - sp->pos.v > gUnivHeight/2)
				shipPos.v -= gUnivHeight;
			if (sp->pos.v - shipPos.v > gUnivHeight/2)
				shipPos.v += gUnivHeight;
			bulletVec.lh = shipPos.h+8 - sp->pos.h;
			bulletVec.lv = shipPos.v+8 - sp->pos.v;			
			scale = 1.0 / sqrt(	(double) bulletVec.lh*bulletVec.lh +
					bulletVec.lv*bulletVec.lv);
			if (scale > 1/300.0) {
				sp->vector.lh = 0;
				sp->vector.lv = 0;
			}
			else {
				bulletVec.lh *= scale * 65536.0;
				bulletVec.lv *= scale * 65536.0;
				bulletVec.lh <<= 2;	// was 2
				bulletVec.lv <<= 2;	// was 2
				sp->vector = bulletVec;
			}
		}
	}
	
	switch (sp->aniState) {
	case 0:
		if (sp->vector.lh == 0L && sp->vector.lv == 0L && MyRandom(3) == 0)
		 	sp->aniState = 82;
		else {
			switch (MyRandom(2)) {
			case 0:	sp->aniState = 1;	break;	// a
			case 1:	sp->aniState = 16;	break;	// b1
			}
		}
		break;
	case 4:
	case 52:
	case 59:
		switch (MyRandom(3)) {
		case 0:	sp->aniState = 5;	break;
		case 1:	sp->aniState = 53;	break;	// a->d
		case 2:	sp->aniState = 60;	break;	// a->d.8
		}
		break;
	case 11:
	case 47:
		switch (MyRandom(2)) {
		case 0:	sp->aniState = 12;	break;	// ->a
		case 1:	sp->aniState = 48;	break;	// ->c
		}
		break;
	case 14:
		sp->aniState = 0;	// a1 -> cycle
		break;
	case 30:
		sp->aniState = 46;	// b1 -> c1
		break;
	case 66:
		sp->aniState = 52;	// d1 -> cycle
		break;
	case 45:
		sp->aniState = 31;	// b2 -> cycle
		break;
	case 81:
		sp->aniState = 67;	// d2 -> cycle
		break;
	case 84:	// Eject
		sp->aniState = 71;
		NewBeater(sp->pos.h,sp->pos.v+16, 0);
		NewBeater(sp->pos.h+4,sp->pos.v+20, 1);
		break;
	default:
		sp->aniState++;
		break;
	}
	SendSpriteStateChange(sp);
}

short TargetSprite(SpriteInstance *sp, short hd, short vd)
{
	double	scale,dist;
	Point	shipPos;
	LongPoint	bulletVec;

	shipPos = gShip->pos;
	if (shipPos.h - sp->pos.h > gUnivWidth/2)
		shipPos.h -= gUnivWidth;
	if (sp->pos.h - shipPos.h > gUnivWidth/2)
		shipPos.h += gUnivWidth;
	if (shipPos.v - sp->pos.v > gUnivHeight/2)
		shipPos.v -= gUnivHeight;
	if (sp->pos.v - shipPos.v > gUnivHeight/2)
		shipPos.v += gUnivHeight;
	bulletVec.lh = shipPos.h+8+hd - sp->pos.h;
	bulletVec.lv = shipPos.v+8+vd - sp->pos.v;	
	dist = sqrt(	(double) bulletVec.lh*bulletVec.lh +
			bulletVec.lv*bulletVec.lv);
	scale = 1.0 / dist;
	bulletVec.lh *= scale * 65536.0;
	bulletVec.lv *= scale * 65536.0;
	bulletVec.lh <<= 3;
	bulletVec.lv <<= 3;
	sp->vector = bulletVec;
	return (short) dist;
}

void NookMove(register SpriteInstance *sp)
{
	double		scale,dist;
	Point		shipPos;
	LongPoint	bulletVec;

	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	shipPos = gShip->pos;
	if (shipPos.h - sp->pos.h > gUnivWidth/2)
		shipPos.h -= gUnivWidth;
	if (sp->pos.h - shipPos.h > gUnivWidth/2)
		shipPos.h += gUnivWidth;
	if (shipPos.v - sp->pos.v > gUnivHeight/2)
		shipPos.v -= gUnivHeight;
	if (sp->pos.v - shipPos.v > gUnivHeight/2)
		shipPos.v += gUnivHeight;

	bulletVec.lh = shipPos.h+8 - sp->pos.h;
	bulletVec.lv = shipPos.v+8 - sp->pos.v;	

	switch (sp->state) {
	case 0:	// Idle a few frames
		if (sp->stateTimer++ > 4)
			sp->state = 1;
		break;
	case 1:		// Go to closest horizontal border point
		sp->stateTimer = 0;	// no break
	case 2:
				// !!! but reduce vertical speed to 0 if within certain range.
		bulletVec.lh -= 100;
		dist = sqrt(	(double) bulletVec.lh*bulletVec.lh +
				bulletVec.lv*bulletVec.lv);
		if (dist < 32)
			sp->state = 2;
		else
			sp->state = 1;
		if (abs(bulletVec.lv) < 80 && abs(bulletVec.lh) > 32)	// Don't cross vertical distance...
			bulletVec.lv = 0;
		scale = 1.0 / dist;
		bulletVec.lh *= scale * 65536.0;
		bulletVec.lv *= scale * 65536.0;
		bulletVec.lh <<= 3;
		bulletVec.lv <<= 3;
		sp->vector = bulletVec;
		if (sp->state == 2) {
			SpriteInstance *spj;
			short	j;
			if (sp->stateTimer++ > 20)
				sp->state = 3;
			for (j = 0,spj=sTable; j < gMaxSprite; ++j,++spj) {
				if (spj->active && spj->type == ST_Cranny && spj->state == 12) {
					sp->state = 3;
					spj->state = 13;
					break;
				}
			}
		}
		break;
	case 3:	// 
		dist = sqrt(	(double) bulletVec.lh*bulletVec.lh +
				bulletVec.lv*bulletVec.lv);
		if (dist > 200)
			sp->state = 1;
		scale = 1.0 / dist;
		bulletVec.lh *= scale * 65536.0;
		bulletVec.lv *= scale * 65536.0;
		bulletVec.lh <<= 3;
		bulletVec.lv <<= 3;
		sp->vector = bulletVec;
		break;
	case 10:	// Idle a few frames
		if (sp->stateTimer++ > 4)
			sp->state = 11;
		break;
	case 11:		// Go to closest horizontal border point
		sp->stateTimer = 0;	// no break
	case 12:
		bulletVec.lh += 100;
		dist = sqrt(	(double) bulletVec.lh*bulletVec.lh +
				bulletVec.lv*bulletVec.lv);
		if (dist < 32)
			sp->state = 12;
		else
			sp->state = 11;
		if (abs(bulletVec.lv) < 80 && abs(bulletVec.lh) > 32)	// Don't cross vertical distance...
			bulletVec.lv = 0;
		scale = 1.0 / dist;
		bulletVec.lh *= scale * 65536.0;
		bulletVec.lv *= scale * 65536.0;
		bulletVec.lh <<= 3;
		bulletVec.lv <<= 3;
		sp->vector = bulletVec;
		if (sp->state == 12) {
			SpriteInstance *spj;
			short	j;
			if (sp->stateTimer++ > 20)
				sp->state = 13;
			for (j = 0,spj=sTable; j < gMaxSprite; ++j,++spj) {
				if (spj->active && spj->type == ST_Nook && spj->state == 2) {
					sp->state = 13;
					spj->state = 3;
					break;
				}
			}
		}
		break;
	case 13:	// 
		dist = sqrt(	(double) bulletVec.lh*bulletVec.lh +
				bulletVec.lv*bulletVec.lv);
		if (dist > 200)
			sp->state = 11;
		scale = 1.0 / dist;
		bulletVec.lh *= scale * 65536.0;
		bulletVec.lv *= scale * 65536.0;
		bulletVec.lh <<= 3;
		bulletVec.lv <<= 3;
		sp->vector = bulletVec;
		break;
	}

	// Update Animation Frame
	sp->tickCtr++;
	if (sp->tickCtr >= sp->aniSpeed) {
		sp->tickCtr = 0;
		sp->aniState++;
		if (sp->aniState >= sDef[sp->type].nbrIcons)
			sp->aniState = 0;
	}
	SendSpriteStateChange(sp);
}

void EMuffinMove(register SpriteInstance *sp)
{
	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	// Update Animation Frame
	sp->tickCtr++;
	if (sp->tickCtr >= sp->aniSpeed) {
		sp->tickCtr = 0;
		sp->aniState++;
		if (sp->aniState >= sDef[sp->type].nbrIcons)
			sp->aniState = 0;
	}
	if (sp->state == 0 && sp->aniState == 4) {
		sp->type = ST_Nook;
		sp->aniState = 2;
		sp->aniSpeed = 2;
		sp->angle = 8;	// Straight Right?
		sp->vector.lh = 0;
		sp->vector.lv = 0;
		sp->state = 0;
		sp->stateTimer = 0;	
	}
	else if (sp->state == 1 && sp->aniState == 0) {
		sp->type = ST_Cranny;
		sp->aniState = 11;
		sp->aniSpeed = 2;
		sp->vector.lh = 0;
		sp->vector.lv = 0;
		sp->state = 10;
		sp->stateTimer = 0;	
	}
	SendSpriteStateChange(sp);
}


void CrannyMove(register SpriteInstance *sp)
{
	NookMove(sp);
}

void ToasterMove(register SpriteInstance *sp)
{
	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	switch (sp->state) {
	case 0:	// Look for nearby teapot
		if (sp->aniState == 0) {
			Point	shipPos,dVec;
			shipPos = gShip->pos;
			if (shipPos.h - sp->pos.h > gUnivWidth/2)
				shipPos.h -= gUnivWidth;
			if (sp->pos.h - shipPos.h > gUnivWidth/2)
				shipPos.h += gUnivWidth;
			if (shipPos.v - sp->pos.v > gUnivHeight/2)
				shipPos.v -= gUnivHeight;
			if (sp->pos.v - shipPos.v > gUnivHeight/2)
				shipPos.v += gUnivHeight;
			dVec.h = shipPos.h - sp->pos.h;
			dVec.v = shipPos.v - sp->pos.v;			
			if (abs(dVec.h) < gVisPlayWidth/2 &&
				abs(dVec.v) < gVisPlayHeight/2) {
				sp->state = 1;
				sp->stateTimer = 0;
				sp->vector.lh = 0L;
				sp->vector.lv = 0L;
				LaunchNookandCranny(sp);
				// gDebug = 1;
				SendSpriteStateChange(sp);
			}
		}
		break;
	case 1: // Unload Muffins
		if (sp->stateTimer++ > 20) {
			sp->state = 2;
			sp->stateTimer = 0L;
			sp->vector.lh = vecTable[sp->angle].lh << 4;
			sp->vector.lv = vecTable[sp->angle].lv << 4;
			SendSpriteStateChange(sp);
		}
		break;
	case 2:
		if (sp->stateTimer++ > 100) {
			// !!! find closest ship...
			Point	shipPos,dVec;
			shipPos = gShip->pos;
			if (shipPos.h - sp->pos.h > gUnivWidth/2)
				shipPos.h -= gUnivWidth;
			if (sp->pos.h - shipPos.h > gUnivWidth/2)
				shipPos.h += gUnivWidth;
			if (shipPos.v - sp->pos.v > gUnivHeight/2)
				shipPos.v -= gUnivHeight;
			if (sp->pos.v - shipPos.v > gUnivHeight/2)
				shipPos.v += gUnivHeight;
			dVec.h = shipPos.h - sp->pos.h;
			dVec.v = shipPos.v - sp->pos.v;			
			if (abs(dVec.h) < gVisPlayWidth &&
				abs(dVec.v) < gVisPlayHeight) {
				NewExplosion(sp);
				ExplodeSprite(sp,LargeSparkLifeSpan);
				NewQuake();
				--gToasterCount;
				return;
			}
			else {
				KillNetSprite(sp);
				--gToasterCount;
				return;
			}
		}
	}
	// Update Animation Frame
	sp->tickCtr++;
	if (sp->tickCtr >= sp->aniSpeed) {
		sp->tickCtr = 0;
		if (++sp->aniState >= sDef[sp->type].nbrIcons)
			sp->aniState = 0;
	}
}


void CubeMove(register SpriteInstance *sp)
{
	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	if (--sp->lifeSpan == 0) {
		NewExplosion(sp);
		ExplodeSprite(sp,LargeSparkLifeSpan);
		NewQuake();
		return;
	}

	// Update Animation Frame
	sp->tickCtr++;
	if (sp->tickCtr >= sp->aniSpeed) {
		sp->tickCtr = 0;
		if (++sp->aniState >= sDef[sp->type].nbrIcons)
			sp->aniState = 0;
		if ((sp->aniState & 1)) {
			LongPoint	bulletVec;
			if (gShip && gShip->type == ST_Teapot || gShip->type == ST_TeapotT) {
				double scale;
				Point	shipPos;
				shipPos = gShip->pos;
				if (shipPos.h - sp->pos.h > gUnivWidth/2)
					shipPos.h -= gUnivWidth;
				if (sp->pos.h - shipPos.h > gUnivWidth/2)
					shipPos.h += gUnivWidth;
				if (shipPos.v - sp->pos.v > gUnivHeight/2)
					shipPos.v -= gUnivHeight;
				if (sp->pos.v - shipPos.v > gUnivHeight/2)
					shipPos.v += gUnivHeight;
				bulletVec.lh = shipPos.h+8 - sp->pos.h;
				bulletVec.lv = shipPos.v+8 - sp->pos.v;			
				scale = 1.0 / sqrt(	(double) bulletVec.lh*bulletVec.lh +
						bulletVec.lv*bulletVec.lv);
				bulletVec.lh *= scale * 65536.0;
				bulletVec.lv *= scale * 65536.0;
				bulletVec.lh <<= 2;
				bulletVec.lv <<= 2;

				sp->vector = bulletVec;
			}
		}
		if (sp->aniState == 9) {
				Point	shipPos,dVec;
				shipPos = gShip->pos;
				if (shipPos.h - sp->pos.h > gUnivWidth/2)
					shipPos.h -= gUnivWidth;
				if (sp->pos.h - shipPos.h > gUnivWidth/2)
					shipPos.h += gUnivWidth;
				if (shipPos.v - sp->pos.v > gUnivHeight/2)
					shipPos.v -= gUnivHeight;
				if (sp->pos.v - shipPos.v > gUnivHeight/2)
					shipPos.v += gUnivHeight;
				dVec.h = shipPos.h - sp->pos.h;
				dVec.v = shipPos.v - sp->pos.v;			
				if (abs(dVec.h) < gVisPlayWidth &&
					abs(dVec.v) < gVisPlayHeight) {
					LaunchBullet(ST_Photon,
								 sp->pos.h+16,sp->pos.v+16,
								 sp->vector.lh << 2,sp->vector.lv << 2,
								 PhotonLifeSpan,
								 BF_BadBulletFlag);
	
					PlaySound(S_EnemyFires, 2);
				}
		}
		SendSpriteStateChange(sp);
	}
}

Boolean ShipCollide(SpriteInstance *sp, SpriteInstance *spj)
{
	register unsigned char *sprPtr;
	register short	j,k;
	Rect			r;
	Point			tPos = spj->pos,bPos = sp->pos;
	Boolean			hitFlag=false;
	unsigned char	typ,cnt;
	LSprite			*spH;

	r.left = max(bPos.h,tPos.h);
	r.top = max(bPos.v,tPos.v);
	r.right = min(bPos.h+sp->width,tPos.h+spj->width);
	r.bottom = min(bPos.v+sp->width,tPos.v+spj->width);

	// Check if collision area has opaque enemy sprite pixels
	spH = (LSprite *) sDef[spj->type].colorMaps;
	sprPtr = &spH->data[spH->frameOffset[spj->aniState]];

	for (j = 0; j+tPos.v < r.top; ++j)
		sprPtr += *((short *) sprPtr);

	for (; j < spj->width && j+tPos.v < r.bottom; ++j) {
		sprPtr += sizeof(short);
		k = 0;
		while (k < spj->width) {
			typ = *sprPtr & LSPR_HiBits;
			cnt = *sprPtr & LSPR_LowBits;
			++sprPtr;
			switch (typ) {
			case LSPR_Mask:
				k += cnt;
				break;
			case LSPR_Opaque:
				sprPtr += cnt;
				while (cnt--) {
					if (k+tPos.h >= r.left && k+tPos.h < r.right)
						goto NextCheck;
					++k;
				}
				break;
			}
		}
	}
	return false;
NextCheck:

	// Check if collision area has opaque ship pixels
	spH = (LSprite *) sDef[sp->type].colorMaps;
	sprPtr = &spH->data[spH->frameOffset[sp->aniState]];

	for (j = 0; j+bPos.v < r.top; ++j)
		sprPtr += *((short *) sprPtr);

	for (; j < sp->width && j+bPos.v < r.bottom; ++j) {
		sprPtr += sizeof(short);
		k = 0;
		while (k < sp->width) {
			typ = *sprPtr & LSPR_HiBits;
			cnt = *sprPtr & LSPR_LowBits;
			++sprPtr;
			switch (typ) {
			case LSPR_Mask:
				k += cnt;
				break;
			case LSPR_Opaque:
				sprPtr += cnt;
				while (cnt--) {
					if (k+bPos.h >= r.left && k+bPos.h < r.right)
						return true;
					++k;
				}
				break;
			}
		}
	}
	return false;
}

Boolean ProcessShipCollide(SpriteInstance *sp, SpriteInstance *spj)
{
	if (!(sp->state & SF_Shield) &&
		spj->type != ST_Yummies &&
		spj->type != ST_Creamer) {
		NewExplosion(sp);
		ExplodeSprite(sp,LargeSparkLifeSpan);
		NewShip();
		return true;
	}

	switch (spj->type) {
	case ST_Wheel:
		NewSubAsteroid(spj,ST_WheelR,-2);
		NewSubAsteroid(spj,ST_WheelG,0);
		NewSubAsteroid(spj,ST_WheelB,2);
		--gAsteroidCnt;
		KillNetSprite(spj);
		PlaySound(S_BreakRings, 2);
		break;
	case ST_Aster:
		NewSubAsteroid(spj,ST_Aster1,-2);
		NewSubAsteroid(spj,ST_Aster2,2);
		--gAsteroidCnt;
		KillNetSprite(spj);
		PlaySound(S_CompoundFracture, 2);
		break;
	case ST_Teacup:
		NewSubAsteroid(spj,ST_SmallCup,-2);
		NewSubAsteroid(spj,ST_SmallCup,0);
		NewSubAsteroid(spj,ST_SmallCup,2);
		--gAsteroidCnt;
		KillNetSprite(spj);
		PlaySound(S_BreakCups, 2);
		break;
	case ST_BigJim:
		NewSubAsteroid(spj,ST_Jim,-2);
		NewSubAsteroid(spj,ST_Jim,0);
		NewSubAsteroid(spj,ST_Jim,2);
		--gAsteroidCnt;
		KillNetSprite(spj);
		PlaySound(S_JimOuch, 3);
		break;
	case ST_WheelR:
	case ST_WheelG:
	case ST_WheelB:
	case ST_Aster1:
	case ST_Aster2:
	case ST_SmallCup:
	case ST_Jim:
		--gAsteroidCnt;
		ExplodeSprite(spj,SmallSparkLifeSpan);
		break;
	case ST_Yummies:
	case ST_Creamer:
		ConsumeYummy(spj);
		break;
	case ST_Saucer:
	case ST_Croissant:
	case ST_Cube:
	case ST_CoffeePot:
	case ST_Beater:
		ExplodeSprite(spj,MedSparkLifeSpan);
		break;
	case ST_Mixer:
		ExplodeSprite(spj,MedSparkLifeSpan);
		--gMixerCount;
		break;
	case ST_Nook:
	case ST_Cranny:
		ExplodeSprite(spj,MedSparkLifeSpan);
		break;
	}
	return false;
}

#define MaxShipSpeed	0x00200000L
#define BulletSpeed		3
#define Debounce		8	// 10
#define UziDebounce		1

short	gFireDelay=0;

void ShipMove(register SpriteInstance *sp)
{
	register short j;
	register SpriteInstance *spj;
	register Point bPos,tPos;

	static Point firingPos[MaxAngles/2] = 
	   {6,24,	7,32,	10,38,	13,42,		
	    19,45,	25,45,	31,41,	34,35,
		37,24,	34,13,	29,5,	25,3,
		19,2,	13,5,	9,11,	7,16};
		

	// Read Keyboard Flags and Set Ship State
	sp->state = 0;
	// Apply KeyFlags
	if (gKeyFlags & KF_Thrust) {
		sp->state |= SF_Thrust;
	}
	if (gKeyFlags & KF_Fire) {
		if (gFireDelay == 0) {
			sp->state |= SF_Fire;
			gFireDelay = ((gShipMode & SM_Uzi)? UziDebounce : Debounce);
		}
		else
			--gFireDelay;
	}
	else if (gFireDelay)
		--gFireDelay;

	if (gKeyFlags & KF_Left)
		sp->state |= SF_Left;
	if (gKeyFlags & KF_Right)
		sp->state |= SF_Right;
	if (gKeyFlags & KF_Shield){
		if (gShieldPower > 0) {
			gShieldPower--;
			sp->state |= SF_Shield;
		}
	}
	else {
		if ((gGameClock & 0xFF) == 0) {		// Increase Shield Power every 16 secs
			if (gShieldPower < MaxShieldPower)
				gShieldPower += ShieldIncrease;
			if (gShieldPower > MaxShieldPower)
				gShieldPower = MaxShieldPower;
		}
	}
	if (gShipMode & SM_AutoShield) {
		if (gShip && gShip->tickCtr > 40)
			gShipMode &= ~SM_AutoShield;
		else
			sp->state |= SF_Shield;
	}		

	// Check for Collision
	bPos = sp->pos;
	for (j = 0,spj=sTable; j < gMaxSprite; ++j,++spj) {
		tPos = spj->pos;
		if (spj->active &&							// Active
			bPos.h+sp->width > tPos.h &&			// In Range
			bPos.v+sp->width > tPos.v &&
			bPos.h < tPos.h+spj->width &&
			bPos.v < tPos.v+spj->width &&
			(gBulletFlag[spj->type] & BF_ShipCollideFlag) > 0)
		{
			if (ShipCollide(sp,spj)) {
				if (ProcessShipCollide(sp,spj))
					return;
			}
		}
	}

	if (sp->state & SF_Left) {
		sp->angle -= 2;
		if (sp->angle < 0)
			sp->angle = MaxAngles-2;
	}
	else if (sp->state & SF_Right) {
		sp->angle += 2;
		if (sp->angle >= MaxAngles)
			sp->angle = 0;
	}

	// Update Animation Frame
	sp->aniState = sp->angle >> 1;

	if (sp->state & SF_Thrust) {
		LongPoint	nv;
		nv.lh = sp->vector.lh + vecTable[sp->angle].lh;
		nv.lv = sp->vector.lv + vecTable[sp->angle].lv;

		if (nv.lh > MaxShipSpeed)	nv.lh = MaxShipSpeed;
		if (nv.lh < -MaxShipSpeed)	nv.lh = -MaxShipSpeed;
		if (nv.lv > MaxShipSpeed)	nv.lv = MaxShipSpeed;
		if (nv.lv < -MaxShipSpeed)	nv.lv = -MaxShipSpeed;
		sp->vector = nv;
		if (sp->type != ST_TeapotT) {
			sp->type = ST_TeapotT;
			PlaySound(S_Thrust, 1);
		}
	}
	else if (sp->state & SF_Shield) {
		if (sp->type != ST_TeapotS) {
			sp->type = ST_TeapotS;
			PlaySound(S_Shield, 1);
		}
	}
	else
		sp->type = ST_Teapot;
	
	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);

	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	tPos = sp->pos;
	tPos.h += (gUnivWidth - gOffsetPt.h); tPos.h %= gUnivWidth;
	tPos.v += (gUnivHeight - gOffsetPt.v); tPos.v %= gUnivHeight;

	// Everything is erased - move scroll offset
	gOffsetPt.h += (sp->vector.lh >> 16);
	gOffsetPt.v += (sp->vector.lv >> 16);
	if (gOffsetPt.h < 0)					gOffsetPt.h += gUnivWidth;
	if (gOffsetPt.v < 0)					gOffsetPt.v += gUnivHeight;
	if (gOffsetPt.h >= gUnivWidth)			gOffsetPt.h -= gUnivWidth;
	if (gOffsetPt.v >= gUnivHeight)			gOffsetPt.v -= gUnivHeight;

	// Moved Firing code down here so bullet origin appears correct...
	if (sp->state & SF_Fire) {
		bPos = sp->pos;
		LaunchBullet(	ST_Bullet,
						bPos.h+firingPos[sp->aniState].h,
						bPos.v+firingPos[sp->aniState].v,
						sp->vector.lh + (vecTable[sp->angle].lh << BulletSpeed),
						sp->vector.lv + (vecTable[sp->angle].lv << BulletSpeed),
						BulletLifeSpan,
						BF_GoodBulletFlag);
		if (gShipMode & SM_Triple) {
			LaunchBullet(	ST_Bullet,
							bPos.h+firingPos[sp->aniState].h,
							bPos.v+firingPos[sp->aniState].v,
							sp->vector.lh + (vecTable[(sp->angle+MaxAngles-1)%MaxAngles].lh << BulletSpeed),
							sp->vector.lv + (vecTable[(sp->angle+MaxAngles-1)%MaxAngles].lv << BulletSpeed),
							BulletLifeSpan,
							BF_GoodBulletFlag);
			LaunchBullet(	ST_Bullet,
							bPos.h+firingPos[sp->aniState].h,
							bPos.v+firingPos[sp->aniState].v,
							sp->vector.lh + (vecTable[(sp->angle+MaxAngles+1)%MaxAngles].lh << BulletSpeed),
							sp->vector.lv + (vecTable[(sp->angle+MaxAngles+1)%MaxAngles].lv << BulletSpeed),
							BulletLifeSpan,
							BF_GoodBulletFlag);
		}
		PlaySound(S_Firing, 1);
	}
	SendSpriteStateChange(sp);

	sp->tickCtr++;
	sp->state &= ~SF_Fire;
}

Boolean BulletCollide(SpriteInstance *sp, SpriteInstance *spj)
{
	register unsigned char *sprPtr;
	register short	j,k;
	Point			tPos = spj->pos,bPos = sp->pos;
	LSprite			*spH;
	unsigned char	typ,cnt;

	// deal with it later...
	// Check if bullet collided with sprite spj
	spH = (LSprite *) sDef[spj->type].colorMaps;
	sprPtr = &spH->data[spH->frameOffset[spj->aniState]];

	for (j = 0; j+tPos.v < bPos.v; ++j)
		sprPtr += *((short *) sprPtr);

	for (; j < spj->width && j+tPos.v < bPos.v+sp->width; ++j) {
		sprPtr += sizeof(short);
		k = 0;
		while (k < spj->width) {
			typ = *sprPtr & LSPR_HiBits;
			cnt = *sprPtr & LSPR_LowBits;
			++sprPtr;
			switch (typ) {
			case LSPR_Mask:
				k += cnt;
				break;
			case LSPR_Opaque:
				sprPtr += cnt;
				while (cnt--) {
					if (k+tPos.h >= bPos.h && k+tPos.h < bPos.h+sp->width)
						return true;
					++k;
				}
				break;
			}
		}
	}
	return false;
}

void ProcessBulletCollide(SpriteInstance *sp, SpriteInstance *spj)
{
	KillNetSprite(sp);		// Kill Bullet

	switch (spj->type) {
	case ST_Wheel:
		NewSubAsteroid(spj,ST_WheelR,-2);
		NewSubAsteroid(spj,ST_WheelG,0);
		NewSubAsteroid(spj,ST_WheelB,2);
		--gAsteroidCnt;
		KillNetSprite(spj);
		AddScore(AsterPoints);
		PlaySound(S_BreakRings, 2);
		break;
	case ST_Aster:
		NewSubAsteroid(spj,ST_Aster1,-2);
		NewSubAsteroid(spj,ST_Aster2,2);
		--gAsteroidCnt;
		KillNetSprite(spj);
		AddScore(AsterPoints);
		PlaySound(S_CompoundFracture, 2);
		break;
	case ST_Teacup:
		NewSubAsteroid(spj,ST_SmallCup,-2);
		NewSubAsteroid(spj,ST_SmallCup,0);
		NewSubAsteroid(spj,ST_SmallCup,2);
		--gAsteroidCnt;
		KillNetSprite(spj);
		AddScore(AsterPoints);
		PlaySound(S_BreakCups, 2);
		break;
	case ST_BigJim:
		NewSubAsteroid(spj,ST_Jim,-2);
		NewSubAsteroid(spj,ST_Jim,0);
		NewSubAsteroid(spj,ST_Jim,2);
		--gAsteroidCnt;
		KillNetSprite(spj);
		AddScore(AsterPoints);
		PlaySound(S_JimOuch, 2);
		break;
	case ST_WheelR:
	case ST_WheelG:
	case ST_WheelB:
	case ST_Aster1:
	case ST_Aster2:
	case ST_SmallCup:
	case ST_Jim:
		--gAsteroidCnt;
		ExplodeSprite(spj,SmallSparkLifeSpan);
		AddScore(AsterPoints);
		break;
	case ST_Saucer:
		if (++spj->hitsTaken >= spj->maxHits)	{		// Max Hits?
			ExplodeSprite(spj,MedSparkLifeSpan);
			AddScore(SaucerPoints);
		}
		break;
	case ST_Croissant:
		if (++spj->hitsTaken >= spj->maxHits)	{		// Max Hits?
			ExplodeSprite(spj,MedSparkLifeSpan);
			AddScore(CroissantPoints);
		}
		break;
	case ST_Cube:
		if (++spj->hitsTaken >= spj->maxHits)	{		// Max Hits?
			ExplodeSprite(spj,MedSparkLifeSpan);
			AddScore(CubePoints);
		}
		break;
	case ST_CoffeePot:
		if (++spj->hitsTaken >= spj->maxHits)	{		// Max Hits?
			ExplodeSprite(spj,MedSparkLifeSpan);
			AddScore(CoffeePotPoints);
		}
		break;
	case ST_Mixer:
		if (++spj->hitsTaken >= spj->maxHits)	{		// Max Hits?
			ExplodeSprite(spj,MedSparkLifeSpan);
			AddScore(MixerPoints);
			--gMixerCount;
		}
		break;
	case ST_Beater:
		ExplodeSprite(spj,MedSparkLifeSpan);
		AddScore(BeaterPoints);
		break;
	case ST_Teapot:
	case ST_TeapotT:
		NewExplosion(sp);
		ExplodeSprite(spj,LargeSparkLifeSpan);
		NewShip();
		break;
	case ST_Yummies:
	case ST_Creamer:
		ExplodeSprite(spj,MedSparkLifeSpan);	// No Points for Exploded Yummies
		break;
	case ST_TeapotS:	// Nada - Shields are up
		PlaySound(S_DudBullet, 2);
		break;
	case ST_Toaster:
		PlaySound(S_DudBullet, 2);
		break;
	case ST_Nook:
	case ST_Cranny:
		ExplodeSprite(spj,MedSparkLifeSpan);
		AddScore(MuffinPoints);
		break;
	}
	return;
}

void BulletMove(register SpriteInstance *sp)
{
	register short j,n;
	register SpriteInstance *spj;
	register Point	bPos,tPos;

	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	// Update Animation Frame
	if (--sp->lifeSpan == 0) {
		KillNetSprite(sp);
		return;
	}
	// Check for Collision
	bPos = sp->pos;
	n = gMaxSprite;
	for (j = 0,spj=sTable; j < n; ++j,++spj) {
		tPos = spj->pos;
		if (spj->active &&							// Active
			// abs((bPos.h >> 5) - (tPos.h >> 5)) < 2 &&	// Range Check
			// abs((bPos.v >> 5) - (tPos.v >> 5)) < 2 &&	// Range Check
			bPos.h+sp->width > tPos.h &&				// In Range
			bPos.v+sp->width > tPos.v &&
			bPos.h < tPos.h+spj->width &&
			bPos.v < tPos.v+spj->width &&
			(gBulletFlag[spj->type] & sp->targetFlags) > 0)
		{
			if (BulletCollide(sp,spj)) {
				ProcessBulletCollide(sp,spj);
				break;
			}
		}
	}
}

void CreamBulletMove(register SpriteInstance *sp)
{
	switch (sp->state) {
	case 0:
		++sp->aniState;
		if (sp->aniState >= 3) {
			sp->aniState = 2;
			sp->state = 1;
		}
		break;
	default:
		--sp->aniState;
		if (sp->aniState < 0) {
			sp->aniState = 2;
			sp->state = 0;
		}
		break;
	}
	BulletMove(sp);
}

void YummyMove(register SpriteInstance *sp)
{
	if (--sp->lifeSpan == 0) {
		KillNetSprite(sp);
		--gYummyCnt;
	}
}

void CreamerMove(register SpriteInstance *sp)
{
	sp->tickCtr++;
	if (sp->tickCtr >= sp->aniSpeed) {
		sp->tickCtr = 0;
		if (++sp->aniState >= sDef[sp->type].nbrIcons)
			sp->aniState = 0;
	}
	if (--sp->lifeSpan == 0) {
		KillNetSprite(sp);
		--gYummyCnt;
	}
}


void SparkMove(register SpriteInstance *sp)
{

	// Update position
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	// Update Animation Frame
	if (--sp->lifeSpan == 0) {
		--gSparkCnt;
		KillSprite(sp);	// Sparks aren't net sprites!!
		return;
	}
}

// This gets swapped into the status sprite's move function, which is
// normally null
// it causes a temporary "spacequake" jitter effect, possibly causing
// collisions with nearby objects
//
void QuakeMove(register SpriteInstance *sp)
{
	register short j,n,i;
	register SpriteInstance *spj;
	static char xOffset[] = {-2,3, 0,-4, 0,-1,2,2};
	static char yOffset[] = {-4,0,-1, 2,-2, 3,2,0};
	if (--sp->lifeSpan == 0) {
		KillNetSprite(sp);
		return;
	}
	n = gMaxSprite;
	i = sp->lifeSpan & 7;
	for (j = 0,spj=sTable; j < n; ++j,++spj) {
		spj->pos.h += xOffset[i];
		spj->pos.v += yOffset[i];
		++i;
		i &= 7;
	}
}

void ExplosionMove(register SpriteInstance *sp)
{
	sp->pos.h += (sp->vector.lh >> 16);
	sp->pos.v += (sp->vector.lv >> 16);
	if (sp->pos.h < 0)
		sp->pos.h += gUnivWidth;
	if (sp->pos.v < 0)
		sp->pos.v += gUnivHeight;
	if (sp->pos.h >= gUnivWidth)
		sp->pos.h -= gUnivWidth;
	if (sp->pos.v >= gUnivHeight)
		sp->pos.v -= gUnivHeight;

	sp->tickCtr++;
	if (sp->tickCtr >= sp->aniSpeed) {
		sp->tickCtr = 0;
		if (++sp->aniState >= sDef[sp->type].nbrIcons)
			KillNetSprite(sp);
	}
}

void NullFunc(register SpriteInstance *sp)
{
}


